# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: featurestore/core/OnlineApi.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='featurestore/core/OnlineApi.proto',
  package='featurestore.core',
  syntax='proto3',
  serialized_options=b'\n\030ai.h2o.featurestore.coreB\016OnlineApiProto',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n!featurestore/core/OnlineApi.proto\x12\x11\x66\x65\x61turestore.core\"Y\n\x15OnlineRetrieveRequest\x12\x33\n\x06header\x18\x01 \x01(\x0b\x32#.featurestore.core.FeatureSetHeader\x12\x0b\n\x03key\x18\x02 \x01(\t\"%\n\x16OnlineRetrieveResponse\x12\x0b\n\x03row\x18\x01 \x01(\t\"X\n\x13OnlineIngestRequest\x12\x33\n\x06header\x18\x01 \x01(\x0b\x32#.featurestore.core.FeatureSetHeader\x12\x0c\n\x04rows\x18\x02 \x03(\t\"[\n\x10\x46\x65\x61tureSetHeader\x12\x12\n\nproject_id\x18\x01 \x01(\t\x12\x16\n\x0e\x66\x65\x61ture_set_id\x18\x02 \x01(\t\x12\x1b\n\x13\x66\x65\x61ture_set_version\x18\x03 \x01(\tB*\n\x18\x61i.h2o.featurestore.coreB\x0eOnlineApiProtob\x06proto3'
)




_ONLINERETRIEVEREQUEST = _descriptor.Descriptor(
  name='OnlineRetrieveRequest',
  full_name='featurestore.core.OnlineRetrieveRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='header', full_name='featurestore.core.OnlineRetrieveRequest.header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='key', full_name='featurestore.core.OnlineRetrieveRequest.key', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=56,
  serialized_end=145,
)


_ONLINERETRIEVERESPONSE = _descriptor.Descriptor(
  name='OnlineRetrieveResponse',
  full_name='featurestore.core.OnlineRetrieveResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='row', full_name='featurestore.core.OnlineRetrieveResponse.row', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=147,
  serialized_end=184,
)


_ONLINEINGESTREQUEST = _descriptor.Descriptor(
  name='OnlineIngestRequest',
  full_name='featurestore.core.OnlineIngestRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='header', full_name='featurestore.core.OnlineIngestRequest.header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='rows', full_name='featurestore.core.OnlineIngestRequest.rows', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=186,
  serialized_end=274,
)


_FEATURESETHEADER = _descriptor.Descriptor(
  name='FeatureSetHeader',
  full_name='featurestore.core.FeatureSetHeader',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='project_id', full_name='featurestore.core.FeatureSetHeader.project_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='feature_set_id', full_name='featurestore.core.FeatureSetHeader.feature_set_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='feature_set_version', full_name='featurestore.core.FeatureSetHeader.feature_set_version', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=276,
  serialized_end=367,
)

_ONLINERETRIEVEREQUEST.fields_by_name['header'].message_type = _FEATURESETHEADER
_ONLINEINGESTREQUEST.fields_by_name['header'].message_type = _FEATURESETHEADER
DESCRIPTOR.message_types_by_name['OnlineRetrieveRequest'] = _ONLINERETRIEVEREQUEST
DESCRIPTOR.message_types_by_name['OnlineRetrieveResponse'] = _ONLINERETRIEVERESPONSE
DESCRIPTOR.message_types_by_name['OnlineIngestRequest'] = _ONLINEINGESTREQUEST
DESCRIPTOR.message_types_by_name['FeatureSetHeader'] = _FEATURESETHEADER
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

OnlineRetrieveRequest = _reflection.GeneratedProtocolMessageType('OnlineRetrieveRequest', (_message.Message,), {
  'DESCRIPTOR' : _ONLINERETRIEVEREQUEST,
  '__module__' : 'featurestore.core.OnlineApi_pb2'
  # @@protoc_insertion_point(class_scope:featurestore.core.OnlineRetrieveRequest)
  })
_sym_db.RegisterMessage(OnlineRetrieveRequest)

OnlineRetrieveResponse = _reflection.GeneratedProtocolMessageType('OnlineRetrieveResponse', (_message.Message,), {
  'DESCRIPTOR' : _ONLINERETRIEVERESPONSE,
  '__module__' : 'featurestore.core.OnlineApi_pb2'
  # @@protoc_insertion_point(class_scope:featurestore.core.OnlineRetrieveResponse)
  })
_sym_db.RegisterMessage(OnlineRetrieveResponse)

OnlineIngestRequest = _reflection.GeneratedProtocolMessageType('OnlineIngestRequest', (_message.Message,), {
  'DESCRIPTOR' : _ONLINEINGESTREQUEST,
  '__module__' : 'featurestore.core.OnlineApi_pb2'
  # @@protoc_insertion_point(class_scope:featurestore.core.OnlineIngestRequest)
  })
_sym_db.RegisterMessage(OnlineIngestRequest)

FeatureSetHeader = _reflection.GeneratedProtocolMessageType('FeatureSetHeader', (_message.Message,), {
  'DESCRIPTOR' : _FEATURESETHEADER,
  '__module__' : 'featurestore.core.OnlineApi_pb2'
  # @@protoc_insertion_point(class_scope:featurestore.core.FeatureSetHeader)
  })
_sym_db.RegisterMessage(FeatureSetHeader)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



from ..base import ArrayField, NumericField, StringField


class FeatureProfile(object):
    feature_type = StringField("features.profile.featureType")


class Feature(object):
    name = StringField("features.name")
    description = StringField("features.description")
    importance = NumericField("features.importance")
    data_type = ArrayField("features.dataTypeHierarchicalPaths")
    profile = FeatureProfile



import collections
import json
import time
import uuid

import grpc
import pkg_resources
import requests
from google.protobuf.empty_pb2 import Empty
from grpc import ChannelConnectivity
from requests.packages.urllib3.exceptions import InsecureRequestWarning

import featurestore.core.CoreService_pb2 as pb
from featurestore.core import (
    CoreService_pb2_grpc,
    generic_client_interceptor,
    interactive_console,
)
from featurestore.core.auth import AuthWrapper
from featurestore.core.client_config import ClientConfig
from featurestore.core.collections.jobs import Jobs
from featurestore.core.collections.projects import Projects
from featurestore.core.config import ConfigUtils
from featurestore.core.credentials import CredentialsHelper
from featurestore.core.data_source_wrappers import SparkDataFrame, get_raw_data_location
from featurestore.core.entities.extract_schema_job import ExtractSchemaJob
from featurestore.core.schema import Schema


class Client:
    def __init__(
        self,
        url: str,
        secure: bool = False,
        root_certificates: str = None,
        config: ClientConfig = ClientConfig(),
    ):
        self._client_config = config
        json_config = json.dumps(
            {
                "methodConfig": [
                    {
                        "name": [{"service": "featurestore.core.CoreService"}],
                        "retryPolicy": {
                            "maxAttempts": 5,
                            "initialBackoff": "0.1s",
                            "maxBackoff": "10s",
                            "backoffMultiplier": 2,
                            "retryableStatusCodes": ["UNAVAILABLE"],
                        },
                    }
                ]
            }
        )
        options = [("grpc.service_config", json_config)]
        if secure:
            default_root_certs = pkg_resources.resource_filename(
                "featurestore", "_credentials/roots.pem"
            )
            credentials = self._get_channel_credentials(
                root_certificates if root_certificates else default_root_certs
            )
            channel = grpc.secure_channel(url, credentials, options)
        else:
            channel = grpc.insecure_channel(url, options)

        auth_interceptor = self._create_auth_interceptor()
        self._channel = grpc.intercept_channel(channel, auth_interceptor)
        self._connection_state = None
        self._config = ConfigUtils.collect_properties()

        def on_connectivity_change(value):
            self._connection_state = value
            return

        self._channel.subscribe(on_connectivity_change, try_to_connect=True)
        if config.wait_for_backend:
            while self._connection_state != ChannelConnectivity.READY:
                print("Connecting to the server {} ...".format(url))
                time.sleep(2)
        else:
            print("Connected to server {}".format(url))

        self._stub = CoreService_pb2_grpc.CoreServiceStub(self._channel)
        self.auth = AuthWrapper(self._stub)
        self.projects = Projects(self._stub)
        self.jobs = Jobs(self._stub)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._channel.close()
        return False

    def close(self):
        self._channel.close()

    def get_version(self) -> str:
        request = Empty()
        response = self._stub.GetVersion(request)
        return response.version

    def extract_schema_from_source_async(
        self, raw_data_location, credentials=None
    ) -> ExtractSchemaJob:
        request = pb.StartExtractSchemaJobRequest()
        if isinstance(raw_data_location, SparkDataFrame):
            raw_data_location._write_to_cache(self._stub)
            data_source = raw_data_location._get_cache_location()
        else:
            data_source = get_raw_data_location(raw_data_location)
        request.raw_data_location.CopyFrom(data_source)
        CredentialsHelper.set_credentials(request, data_source, credentials)
        job_id = self._stub.StartExtractSchemaJob(request)
        return ExtractSchemaJob(self._stub, job_id)

    @interactive_console.record_stats
    def extract_schema_from_source(self, raw_data_location, credentials=None) -> Schema:
        job = self.extract_schema_from_source_async(raw_data_location, credentials)
        while not job.is_done():
            job.show_progress()
            time.sleep(2)
        job.show_progress()  # there is possibility that some progress was pushed before finishing job
        return job.get_result()

    def _has_online_retrieve_permissions(self, project_name, feature_set_name):
        request = pb.HasPermissionToRetrieveRequest()
        request.project_name = project_name
        request.feature_set_name = feature_set_name
        response = self._stub.HasPermissionToRetrieve(request)
        return response.has_retrieve_permission

    def online_retrieve(
        self,
        project_name,
        feature_set_name,
        primary_key,
        secondary_key=None,
        verify=False,
    ):
        # this method will be eventually removed, H2O's online is using method onlineRetrieve on feature set entity
        if not self._has_online_retrieve_permissions(project_name, feature_set_name):
            raise Exception(
                "User does not have enough permissions to execute online retrieve call"
            )
        resp = self._stub.GetOnlineRetrieveMeta(Empty())
        if not resp.is_enabled:
            raise Exception(
                "Online feature store integration is not enabled on the backend"
            )
        url = (
            resp.base_online_url
            + requests.utils.quote(project_name)
            + "/"
            + requests.utils.quote(feature_set_name)
            + "/"
            + requests.utils.quote(primary_key)
        )
        if secondary_key:
            params = dict(secondaryKey=requests.utils.quote(secondary_key))
        else:
            params = dict(secondaryKey="")
        session = requests.Session()
        session.auth = (resp.username, resp.password)
        if not verify:
            requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
        response = session.get(url, params=params, verify=verify).json()
        return response

    def show_progress(self, interactive):
        ConfigUtils.set_property(
            self._config, ConfigUtils.INTERACTIVE_LOGGING, str(interactive)
        )

    @staticmethod
    def _get_channel_credentials(cert_location: str) -> grpc.ChannelCredentials:
        if cert_location is not None:
            with open(cert_location, "rb") as cert_file:
                return grpc.ssl_channel_credentials(cert_file.read())
        return grpc.ssl_channel_credentials(None)

    def _create_auth_interceptor(self):
        def intercept_call(
            client_call_details, request_iterator, request_streaming, response_streaming
        ):
            if client_call_details.method in (
                "/featurestore.core.CoreService/Login",
                "/featurestore.core.CoreService/GetAccessToken",
            ):
                return client_call_details, request_iterator, None
            else:
                metadata = []
                if client_call_details.metadata is not None:
                    metadata = list(client_call_details.metadata)
                metadata.append(("authorization", self.auth._obtain_token(),))
                metadata.append(("request-id", str(uuid.uuid4())))
                client_call_details = _ClientCallDetails(
                    client_call_details.method,
                    self._client_config.timeout,
                    metadata,
                    client_call_details.credentials,
                )
                return client_call_details, request_iterator, None

        return generic_client_interceptor.create(intercept_call)


class _ClientCallDetails(
    collections.namedtuple(
        "_ClientCallDetails", ("method", "timeout", "metadata", "credentials")
    ),
    grpc.ClientCallDetails,
):
    pass



import grpc


class _GenericClientInterceptor(
    grpc.UnaryUnaryClientInterceptor,
    grpc.UnaryStreamClientInterceptor,
    grpc.StreamUnaryClientInterceptor,
    grpc.StreamStreamClientInterceptor,
):
    def __init__(self, interceptor_function):
        self._fn = interceptor_function

    def intercept_unary_unary(self, continuation, client_call_details, request):
        new_details, new_request_iterator, postprocess = self._fn(
            client_call_details, iter((request,)), False, False
        )
        response = continuation(new_details, next(new_request_iterator))
        return postprocess(response) if postprocess else response

    def intercept_unary_stream(self, continuation, client_call_details, request):
        new_details, new_request_iterator, postprocess = self._fn(
            client_call_details, iter((request,)), False, True
        )
        response_it = continuation(new_details, next(new_request_iterator))
        return postprocess(response_it) if postprocess else response_it

    def intercept_stream_unary(
        self, continuation, client_call_details, request_iterator
    ):
        new_details, new_request_iterator, postprocess = self._fn(
            client_call_details, request_iterator, True, False
        )
        response = continuation(new_details, new_request_iterator)
        return postprocess(response) if postprocess else response

    def intercept_stream_stream(
        self, continuation, client_call_details, request_iterator
    ):
        new_details, new_request_iterator, postprocess = self._fn(
            client_call_details, request_iterator, True, True
        )
        response_it = continuation(new_details, new_request_iterator)
        return postprocess(response_it) if postprocess else response_it


def create(intercept_call):
    return _GenericClientInterceptor(intercept_call)



from typing import List

from google.protobuf.empty_pb2 import Empty

from .. import CoreService_pb2 as pb
from ..entities.pat import PersonalAccessToken


class PersonalAccessTokens:
    def __init__(self, stub):
        self._stub = stub

    def generate(self, name: str, description: str, expiry_date: str = None) -> str:
        request = pb.GenerateTokenRequest()
        request.name = name
        request.description = description
        if expiry_date:
            request.expiry_date = expiry_date
        response = self._stub.GenerateToken(request)
        return response.token

    def list(self) -> List[PersonalAccessToken]:
        request = Empty()
        response = self._stub.ListTokens(request)
        return [PersonalAccessToken(self._stub, pat) for pat in response.tokens]

    def get(self, token_id: str) -> PersonalAccessToken:
        request = pb.TokenRequest()
        request.token_id = token_id
        response = self._stub.GetToken(request)
        return PersonalAccessToken(self._stub, response.token)

    def __repr__(self):
        return "This class wraps together methods working with Personal Access Tokens (PATs)"



import collections
import time
from datetime import timedelta
from functools import wraps

import featurestore.core.CoreService_pb2 as pb
from featurestore.core.config import ConfigUtils

METRICS_START_TIME = time.time()
Metrics = collections.namedtuple(
    "Metrics", "func_name sum_time num_calls min_time max_time avg_time"
)


def log(message):
    if ConfigUtils.is_interactive_print_enabled():
        print(message)


def record_stats(fn):
    @wraps(fn)
    def with_perf(*args, **kwargs):
        start = time.time()
        ret = fn(*args, **kwargs)
        end = time.time()
        delta = timedelta(seconds=end - start).total_seconds()
        print("\n\nTime taken - %3f seconds" % delta)
        return ret

    return with_perf


class Progress:
    def __init__(self, stub, job):
        self._stub = stub
        self.job = job
        self.next_index = 0

    def show(self):
        if ConfigUtils.is_interactive_print_enabled():
            request = pb.JobProgressInput(
                job_id=self.job.job_id, next_index=self.next_index
            )
            response = self._stub.GetJobProgress(request)
            if response.status:
                self.next_index += len(response.status)
                for status in response.status:
                    log(f"Job ID: {self.job.job_id}, Status: {status}")



from .. import CoreService_pb2 as pb
from ..utils import Utils

CATEGORICAL = pb.FeatureType.Categorical
NUMERICAL = pb.FeatureType.Numerical
TEMPORAL = pb.FeatureType.Temporal
TEXT = pb.FeatureType.Text


class Feature:
    def __init__(self, feature):
        self._feature = feature

    @property
    def name(self):
        return self._feature.name

    @property
    def version(self):
        return self._feature.version

    @property
    def special(self):
        return self._feature.special

    @special.setter
    def special(self, value):
        self._feature.special = value

    @property
    def version_change(self):
        return self._feature.version_change

    @version_change.setter
    def version_change(self, value):
        self._feature.version_change = value

    @property
    def status(self):
        return self._feature.status

    @status.setter
    def status(self, value):
        self._feature.status = value

    @property
    def data_type(self):
        return self._feature.data_type

    @property
    def profile(self):
        return FeatureProfile(self._feature)

    @property
    def description(self):
        return self._feature.description

    @description.setter
    def description(self, value):
        self._feature.description = value

    @property
    def importance(self):
        return self._feature.importance

    @importance.setter
    def importance(self, value):
        self._feature.importance = value

    @property
    def monitoring(self):
        return self._feature.monitoring

    @property
    def marked_for_masking(self):
        return self._feature.marked_for_masking

    def __repr__(self):
        return Utils.pretty_print_proto(self._feature)


class FeatureProfile:
    def __init__(self, feature):
        self._feature = feature
        self._profile = self._feature.profile

    @property
    def feature_type(self):
        return self._profile.feature_type

    @feature_type.setter
    def feature_type(self, value):
        self._profile.feature_type = pb.FeatureType.Name(value).title()

    @property
    def categorical_statistics(self):
        return self._profile.categorical

    @property
    def statistics(self):
        return self._profile.statistics

    def __repr__(self):
        return Utils.pretty_print_proto(self._profile)


class FeatureStatistics:
    def __init__(self, feature):
        self._feature = feature
        self._stats = self._feature.categorical

    @property
    def max(self):
        return self._stats.max

    @property
    def mean(self):
        return self._stats.mean

    @property
    def median(self):
        return self._stats.median

    @property
    def min(self):
        return self._stats.min

    @property
    def stddev(self):
        return self._stats.stddev

    @property
    def stddev_rec_count(self):
        return self._stats.stddev_rec_count

    @property
    def null_count(self):
        return self._stats.null_count

    @property
    def nan_count(self):
        return self._stats.nan_count

    @property
    def unique(self):
        return self._stats.unique

    def __repr__(self):
        return Utils.pretty_print_proto(self._stats)


class CategoricalStatistics:
    def __init__(self, feature):
        self._feature = feature
        self._categorical = self._feature.categorical

    @property
    def unique(self):
        return self._categorical.unique

    @property
    def top(self):
        return [FeatureTop(top) for top in self._categorical.top]

    def __repr__(self):
        return Utils.pretty_print_proto(self._categorical)


class FeatureTop:
    def __init__(self, feature):
        self._feature = feature
        self._top = self._feature.top

    @property
    def name(self):
        return self._top.name

    @property
    def count(self):
        return self._top.count

    def __repr__(self):
        return Utils.pretty_print_proto(self._top)


class Monitoring:
    def __init__(self, feature):
        self._feature = feature
        self._monitoring = self._feature.monitoring

    @property
    def anomaly_detection(self):
        return self._monitoring.anomaly_detection

    @anomaly_detection.setter
    def anomaly_detection(self, value):
        self._monitoring.anomaly_detection = value

    def __repr__(self):
        return Utils.pretty_print_proto(self._monitoring)
